// conin.c, compiler-specific code for MinGW/GCC
// add these linker options: -Wl,-e,__start -nostartfiles

#ifdef _WIN32_WINNT
# undef _WIN32_WINNT
#endif
#define _WIN32_WINNT 0x501
#define WIN32_LEAN_AND_MEAN
#define NOUSER
#define NOSERVICE
#define NOMCX
#define NOIME
#define NOKERNEL
#include <windows.h>
#include <conio.h>
#include <stdlib.h>

typedef struct
{
  int newmode;
} _startupinfo;

typedef struct
{
  HANDLE conIn;
  DWORD ret;
} THREADARG;

extern
#ifdef __cplusplus
  "C"
#endif
int __getmainargs(int *, char ***, char ***, int, _startupinfo *);

DWORD WINAPI GetInput(THREADARG *const p_threadarg);


#ifdef __i386
  void _start(void)
#else
  void __start(void)
#endif
{
  int              argc      = 0;
  char          ** argv      = NULL,
                ** envp      = NULL;
  _startupinfo     si        = {0};
  DWORD            oldMode   = 0u,
                   ret       = 0u,
                   timeout   = INFINITE;
  HANDLE           hthread   = NULL;
  THREADARG        threadarg = {GetStdHandle(STD_INPUT_HANDLE), 0u};

  __getmainargs(&argc, &argv, &envp, 0, &si);
  GetConsoleMode(threadarg.conIn, &oldMode);
  SetConsoleMode(threadarg.conIn, (oldMode & ~ENABLE_QUICK_EDIT_MODE) | ENABLE_MOUSE_INPUT | ENABLE_EXTENDED_FLAGS);

  if (argc > 1)
    timeout = strtoul(argv[1u], NULL, 0);

  hthread = CreateThread(NULL, 0u, (LPTHREAD_START_ROUTINE)GetInput, (LPVOID)&threadarg, 0u, NULL);
  ret = WaitForSingleObject(hthread, timeout);

  if (ret == WAIT_TIMEOUT)
    TerminateThread(hthread, 0u);

  CloseHandle(hthread);
  SetConsoleMode(threadarg.conIn, oldMode | ENABLE_EXTENDED_FLAGS);

  exit(ret == 0u ? (int)threadarg.ret : 0);
}


DWORD WINAPI GetInput(THREADARG *const p_threadarg)
{
  DWORD         cntRead    = 0u;
  INPUT_RECORD  inpRec[4u] = {0};

  for(;;)
  {
    if (_kbhit())
    {
      if ((p_threadarg->ret = _getch()) == 0u || p_threadarg->ret == 224u)
        p_threadarg->ret = 256u + _getch();

      return 0u;
    }
    else
    {
      WaitForSingleObject(p_threadarg->conIn, INFINITE);
      PeekConsoleInput(p_threadarg->conIn, inpRec, 4u, &cntRead);
      for(DWORD i = 0; i < cntRead; ++i)
      {
        if(inpRec[i].EventType == MOUSE_EVENT)
        {
          p_threadarg->ret = (DWORD)(inpRec[i].Event.MouseEvent.dwMousePosition.Y & 0x3FFFu) << 14u | (inpRec[i].Event.MouseEvent.dwMousePosition.X & 0x3FFFu);
          if (inpRec[i].Event.MouseEvent.dwButtonState == FROM_LEFT_1ST_BUTTON_PRESSED || inpRec[i].Event.MouseEvent.dwButtonState == RIGHTMOST_BUTTON_PRESSED)
            p_threadarg->ret |= inpRec[i].Event.MouseEvent.dwButtonState << 29u;
          else if (inpRec[i].Event.MouseEvent.dwEventFlags == MOUSE_MOVED)
            p_threadarg->ret |= 3u << 29u;
          else
          {
            FlushConsoleInputBuffer(p_threadarg->conIn);
            continue;
          }

          FlushConsoleInputBuffer(p_threadarg->conIn);
          return 0u;
        }
      }
    }
  }
  return 0u;
}
